from dataclasses import dataclass

import pandas as pd


@dataclass
class Metadata:
    match_id: int
    pitch_dimensions: list
    periods_frames: pd.DataFrame
    frame_rate: int

    home_team_id: int
    home_team_name: str
    home_players: pd.DataFrame
    home_score: int
    home_formation: str

    away_team_id: int
    away_team_name: str
    away_players: pd.DataFrame
    away_score: int
    away_formation: str

    def __post_init__(self):
        # match id
        if not isinstance(self.match_id, int):
            raise TypeError(
                f"match_id ({self.match_id}) should be an integer, not \
                    {type(self.match_id)}"
            )

        # pitch_dimensions
        if not isinstance(self.pitch_dimensions, list):
            raise TypeError(
                f"pitch_dimensions ({self.pitch_dimensions}) should be a \
                    list, not a {type(self.pitch_dimensions)}"
            )
        if not len(self.pitch_dimensions) == 2:
            raise ValueError(
                f"pitch_dimensions should contain, two values: a length and a width \
                    of the pitch, current input is {self.pitch_dimensions}"
            )
        if not all([isinstance(x, float) for x in self.pitch_dimensions]):
            raise TypeError(
                f"Both values in pitch dimensions should by floats, current inputs \
                    {[type(x) for x in self.pitch_dimensions]}"
            )

        # periods_frames
        if not isinstance(self.periods_frames, pd.DataFrame):
            raise TypeError(
                f"periods_frames should be a pandas dataframe, not a \
                    {type(self.periods_frames)}"
            )
        if "period" not in self.periods_frames.columns:
            raise ValueError("'period' should be one of the columns in period_frames")
        if any(
            [
                x not in self.periods_frames["period"].value_counts().index
                for x in [1, 2, 3, 4, 5]
            ]
        ) or not all(self.periods_frames["period"].value_counts() == 1):

            res = self.periods_frames["period"]
            raise ValueError(
                f"'period' column in period_frames should contain only the values \
                    [1, 2, 3, 4, 5]. Now it's {res}"
            )

        # frame_rate
        if not pd.isnull(self.frame_rate):
            if not isinstance(self.frame_rate, int):
                raise TypeError(
                    f"frame_rate should be an integer, not a {type(self.frame_rate)}"
                )
            if self.frame_rate < 1:
                raise ValueError(
                    f"frame_rate should be positive, not {self.frame_rate}"
                )

        # team id's
        for team, team_id in zip(
            ["home", "away"], [self.home_team_id, self.away_team_id]
        ):
            if not isinstance(team_id, int) and not isinstance(team_id, str):
                raise TypeError(
                    f"{team} team id should be an integer or string, not a \
                        {type(team_id)}"
                )

        # team names
        for team, name in zip(
            ["home", "away"], [self.home_team_name, self.away_team_name]
        ):
            if not isinstance(name, str):
                raise TypeError(
                    f"{team} team name should be a string, not a {type(name)}"
                )

        # team scores
        for team, score in zip(["home", "away"], [self.home_score, self.away_score]):
            if not pd.isnull(score):
                if not isinstance(score, int):
                    raise TypeError(
                        f"{team} team score should be an integer, not a {type(score)}"
                    )
                if score < 0:
                    raise ValueError(
                        f"{team} team score should be positive, not {score}"
                    )

        # team formations
        for team, form in zip(
            ["home", "away"], [self.home_formation, self.away_formation]
        ):
            if not pd.isnull(form):
                if not isinstance(form, str):
                    raise TypeError(
                        f"{team} team formation should be a string, not a {type(form)}"
                    )
                if len(form) > 4:
                    raise ValueError(
                        f"{team} team formation should be of length 4 or smaller, not \
                            {len(form)}"
                    )

        # team players
        for team, players in zip(
            ["home", "away"], [self.home_players, self.away_players]
        ):
            if not isinstance(players, pd.DataFrame):
                raise TypeError(
                    f"{team} team players should be a pandas dataframe, not a \
                        {type(players)}"
                )
            if not all(
                [x in players.columns for x in ["id", "full_name", "shirt_num"]]
            ):
                raise ValueError(
                    f"{team} team players should contain at least the column \
                        ['id', 'full_name', 'shirt_num'], now its {players.columns}"
                )

    def __eq__(self, other):
        result = [
            self.match_id == other.match_id,
            all(
                [s == o for s, o in zip(self.pitch_dimensions, other.pitch_dimensions)]
            ),
            self.periods_frames.equals(other.periods_frames),
            self.frame_rate == other.frame_rate
            if not pd.isnull(self.frame_rate)
            else pd.isnull(other.frame_rate),
            self.home_team_id == other.home_team_id,
            self.home_team_name == other.home_team_name,
            self.home_players.equals(other.home_players),
            self.home_score == other.home_score
            if not pd.isnull(self.home_score)
            else pd.isnull(other.home_score),
            self.home_formation == other.home_formation,
            self.away_team_id == other.away_team_id,
            self.away_team_name == other.away_team_name,
            self.away_players.equals(other.away_players),
            self.away_score == other.away_score
            if not pd.isnull(self.away_score)
            else pd.isnull(other.away_score),
            self.away_formation == other.away_formation,
        ]

        return all(result)

    def copy(self):
        return Metadata(
            match_id=self.match_id,
            pitch_dimensions=list(self.pitch_dimensions),
            periods_frames=self.periods_frames.copy(),
            frame_rate=self.frame_rate,
            home_team_id=self.home_team_id,
            home_team_name=self.home_team_name,
            home_players=self.home_players.copy(),
            home_score=self.home_score,
            home_formation=self.home_formation,
            away_team_id=self.away_team_id,
            away_team_name=self.away_team_name,
            away_players=self.away_players.copy(),
            away_score=self.away_score,
            away_formation=self.away_formation,
        )



import datetime as dt
import os

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup

from databallpy.load_data.metadata import Metadata
from databallpy.utils import _to_float, _to_int


def _get_td_channels(metadata_loc: str, metadata: Metadata) -> pd.DataFrame:
    """Function to get the channels for every timeperiod with what players are
    referred to in the raw tracking data

    Args:
        metadata_loc (str): locatin of the metadata
        metadata (Metadata): the Metadata of the match

    Returns:
        pd.DataFrame: df with for every timestamp what players are referred to in
        the raw tracking data
    """
    if os.path.exists(metadata_loc):
        file = open(metadata_loc, "r")
        lines = file.read()
        file.close()
        soup = BeautifulSoup(lines, "xml")
    else:
        soup = BeautifulSoup(metadata_loc.strip(), "xml")

    res = {"start": [], "end": [], "ids": []}
    for idx, data_format_specification in enumerate(
        soup.find_all("DataFormatSpecification")
    ):
        res["ids"].append([])
        res["start"].append(int(data_format_specification.attrs["startFrame"]))
        res["end"].append(int(data_format_specification.attrs["endFrame"]))
        for channel in data_format_specification.findChildren("PlayerChannelRef"):
            full_name = channel.attrs["playerChannelId"].split("_")[0]
            value = channel.attrs["playerChannelId"].split("_")[1]
            if "y" in value:
                continue
            home_mask = (
                metadata.home_players["full_name"]
                .str.lower()
                .str.replace(" ", "")
                .isin([full_name])
            )
            away_mask = (
                metadata.away_players["full_name"]
                .str.lower()
                .str.replace(" ", "")
                .isin([full_name])
            )
            if home_mask.any():
                team = "home"
                shirt_num = metadata.home_players.loc[home_mask, "shirt_num"].iloc[0]
            else:
                team = "away"
                shirt_num = metadata.away_players.loc[away_mask, "shirt_num"].iloc[0]
            res["ids"][idx].append(f"{team}_{shirt_num}")
    return pd.DataFrame(res)


def _get_metadata(metadata_loc: str) -> Metadata:
    """Function to get the metadata of the match

    Args:
        metadata_loc (str): Location of the metadata .xml file

    Returns:
        Metadata: all information of the match
    """
    if os.path.exists(metadata_loc):
        file = open(metadata_loc, "r")
        lines = file.read()
        file.close()
        soup = BeautifulSoup(lines, "xml")
    else:
        soup = BeautifulSoup(metadata_loc.strip(), "xml")

    match_id = _to_int(soup.find("Session").attrs["id"])
    pitch_size_x = _to_float(soup.find("FieldSize").find("Width").text)
    pitch_size_y = _to_float(soup.find("FieldSize").find("Height").text)
    frame_rate = _to_int(soup.find("FrameRate").text)
    datetime = pd.to_datetime(soup.find("Start").text).tz_localize(None)

    periods_dict = {
        "period": [],
        "start_frame": [],
        "end_frame": [],
        "start_time_td": [],
        "end_time_td": [],
    }

    periods_map = {
        "first_half_start": 1,
        "first_half_end": 1,
        "second_half_start": 2,
        "second_half_end": 2,
        "first_extra_half_start": 3,
        "first_extra_half_end": 3,
        "second_extra_half_start": 4,
        "second_extra_half_end": 4,
    }

    for period_soup in soup.find("ProviderGlobalParameters").find_all(
        "ProviderParameter"
    ):
        name = period_soup.find("Name").text
        period = periods_map[name]

        current_timestamp = _to_int(period_soup.find("Value").text)

        if "start" in name:
            periods_dict["period"].append(period)
            periods_dict["start_frame"].append(current_timestamp)
            first_timestamp = periods_dict["start_frame"][0]
            seconds = (current_timestamp - first_timestamp) / frame_rate
            periods_dict["start_time_td"].append(
                datetime + dt.timedelta(milliseconds=seconds * 1000)
            ) if not current_timestamp == -999 else periods_dict[
                "start_time_td"
            ].append(
                pd.to_datetime("NaT")
            )

        elif "end" in name:
            periods_dict["end_frame"].append(current_timestamp)
            first_timestamp = periods_dict["start_frame"][0]
            seconds = (current_timestamp - first_timestamp) / frame_rate
            periods_dict["end_time_td"].append(
                datetime + dt.timedelta(milliseconds=seconds * 1000)
            ) if not current_timestamp == -999 else periods_dict["end_time_td"].append(
                pd.to_datetime("NaT")
            )

    # add fifth period
    periods_dict["period"].append(5)
    periods_dict["start_frame"].append(-999)
    periods_dict["end_frame"].append(-999)
    periods_dict["start_time_td"].append(pd.to_datetime("NaT"))
    periods_dict["end_time_td"].append(pd.to_datetime("NaT"))

    teams_info = {}
    for team in soup.find_all("Team"):
        team_info = {}
        team_info["team_name"] = team.find("Name").text
        team_info["team_id"] = team.attrs["id"]
        team_info["formation"] = ""
        score = soup.find("Score")
        if score.attrs["idLocalTeam"] == team_info["team_id"]:
            team_info["side"] = "home"
            team_info["score"] = _to_int(score.find("LocalTeamScore").text)
        else:
            team_info["side"] = "away"
            team_info["score"] = _to_int(score.find("VisitingTeamScore").text)

        teams_info[team_info["side"]] = team_info

    players = soup.find_all("Player")
    team_dicts = [
        {
            "team_id": teams_info["home"]["team_id"],
            "player_dict": {
                "id": [],
                "full_name": [],
                "formation_place": [],
                "position": [],
                "starter": [],
                "shirt_num": [],
            },
        },
        {
            "team_id": teams_info["away"]["team_id"],
            "player_dict": {
                "id": [],
                "full_name": [],
                "formation_place": [],
                "position": [],
                "starter": [],
                "shirt_num": [],
            },
        },
    ]

    for player in players:
        for team in team_dicts:
            if player.attrs["teamId"] == team["team_id"]:
                res_dict = team["player_dict"]

                res_dict["id"].append(_to_int(player.attrs["id"][1:]))
                res_dict["full_name"].append(player.find("Name").text)
                res_dict["shirt_num"].append(_to_int(player.find("ShirtNumber").text))
                res_dict["starter"].append(np.nan)
                for param in player.findChildren("ProviderParameter"):
                    if param.find("Name").text == "position_type":
                        res_dict["position"].append(param.find("Value").text)
                    elif param.find("Name").text == "position_index":
                        res_dict["formation_place"].append(
                            _to_int(param.find("Value").text)
                        )

    home_players = team_dicts[0]["player_dict"]
    away_players = team_dicts[1]["player_dict"]

    metadata = Metadata(
        match_id=_to_int(match_id),
        pitch_dimensions=[float(pitch_size_x), float(pitch_size_y)],
        periods_frames=pd.DataFrame(periods_dict),
        frame_rate=_to_int(frame_rate),
        home_team_id=teams_info["home"]["team_id"],
        home_team_name=teams_info["home"]["team_name"],
        home_players=pd.DataFrame(home_players),
        home_score=teams_info["home"]["score"],
        home_formation=teams_info["home"]["formation"],
        away_team_id=teams_info["away"]["team_id"],
        away_team_name=teams_info["away"]["team_name"],
        away_players=pd.DataFrame(away_players),
        away_score=teams_info["away"]["score"],
        away_formation=teams_info["away"]["formation"],
    )
    return metadata


def _update_metadata(td_channels: pd.DataFrame, metadata: Metadata) -> Metadata:
    """Function to add the starters and formation based on the metadata and tracking
    data channels to the metadata

    Args:
        td_channels (pd.DataFrame): tracking data channels, what indexes in the
        tracking data belong to which players, all players in the first line are
        starters
        metadata (Metadata): metadata of the match

    Returns:
        Metadata: updated metadata of the match
    """
    home_formation = [0, 0, 0, 0]
    away_formation = [0, 0, 0, 0]
    metadata.home_players["starter"] = False
    metadata.away_players["starter"] = False
    starting = td_channels.iloc[0]["ids"]

    for starter in starting:
        team, shirt_number = starter.split("_")
        shirt_number = _to_int(shirt_number)
        if team == "home":
            metadata.home_players.loc[
                metadata.home_players["shirt_num"] == shirt_number, "starter"
            ] = True
            formation_place = metadata.home_players.loc[
                metadata.home_players["shirt_num"] == shirt_number, "formation_place"
            ].iloc[0]
            home_formation[formation_place] += 1
        else:
            metadata.away_players.loc[
                metadata.away_players["shirt_num"] == shirt_number, "starter"
            ] = True
            formation_place = metadata.away_players.loc[
                metadata.away_players["shirt_num"] == shirt_number, "formation_place"
            ].iloc[0]
            away_formation[formation_place] += 1

    metadata.home_formation = "".join(str(i) for i in home_formation)
    metadata.away_formation = "".join(str(i) for i in away_formation)

    return metadata



import datetime as dt
import io
import json
import os
from typing import Tuple, Union

import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup

from databallpy.load_data.event_data._normalize_playing_direction_events import (
    _normalize_playing_direction_events,
)
from databallpy.load_data.metadata import Metadata
from databallpy.load_data.metrica_metadata import (
    _get_metadata,
    _get_td_channels,
    _update_metadata,
)
from databallpy.utils import _to_float, _to_int


def load_metrica_event_data(
    event_data_loc: str, metadata_loc: str
) -> Tuple[pd.DataFrame, Metadata]:
    """Function to load the metrica event data.

    Args:
        event_data_loc (str): location of the event data .json file
        metadata_loc (str): location of the metadata .xml file

    Raises:
        TypeError: type error if event_data_loc, or metadata_loc is not a valid input
        type (str)

    Returns:
        Tuple[pd.DataFrame, Metadata]: The event data and the metadata
    """

    if isinstance(event_data_loc, str) and "{" not in event_data_loc:
        assert os.path.exists(event_data_loc)
        assert os.path.exists(metadata_loc)
    elif isinstance(event_data_loc, str) and "{" in event_data_loc:
        pass
    else:
        raise TypeError(
            "tracking_data_loc must be either a str or a StringIO object,"
            f" not a {type(event_data_loc)}"
        )

    metadata = _get_metadata(metadata_loc)
    td_channels = _get_td_channels(metadata_loc, metadata)
    metadata = _update_metadata(td_channels, metadata)
    event_data = _get_event_data(event_data_loc)

    # rescale the event locations, metrica data is scaled between 0 and 1.
    for col in [x for x in event_data.columns if "_x" in x]:
        event_data[col] = (
            event_data[col] * metadata.pitch_dimensions[0]
            - metadata.pitch_dimensions[0] / 2.0
        )
    for col in [x for x in event_data.columns if "_y" in x]:
        event_data[col] = (
            event_data[col] * metadata.pitch_dimensions[1]
            - metadata.pitch_dimensions[1] / 2.0
        )

    # add datetime based on frame numbers
    first_frame = metadata.periods_frames.loc[
        metadata.periods_frames["period"] == 1, "start_frame"
    ].iloc[0]
    start_time = metadata.periods_frames.loc[
        metadata.periods_frames["period"] == 1, "start_time_td"
    ].iloc[0]
    frame_rate = metadata.frame_rate
    rel_timedelta = [
        dt.timedelta(milliseconds=(x - first_frame) / frame_rate * 1000)
        for x in event_data["td_frame"]
    ]
    event_data["datetime"] = [pd.to_datetime(start_time) + x for x in rel_timedelta]

    event_data = _normalize_playing_direction_events(
        event_data, metadata.home_team_id, metadata.away_team_id
    )

    return event_data, metadata


def load_metrica_open_event_data() -> Tuple[pd.DataFrame, Metadata]:

    metadata_link = "https://raw.githubusercontent.com/metrica-sports/sample-data\
        /master/data/Sample_Game_3/Sample_Game_3_metadata.xml"
    ed_link = "https://raw.githubusercontent.com/metrica-sports/sample-data\
        /master/data/Sample_Game_3/Sample_Game_3_events.json"

    print("Downloading Metrica open event data...", end="")
    raw_ed = requests.get(ed_link).text
    raw_metadata = requests.get(metadata_link).text
    print(" Done!")

    return load_metrica_event_data(raw_ed, raw_metadata)


def _get_event_data(event_data_loc: Union[str, io.StringIO]) -> pd.DataFrame:

    if isinstance(event_data_loc, str) and "{" not in event_data_loc:
        file = open(event_data_loc)
        lines = file.readlines()
        file.close()
        raw_data = "".join(str(i) for i in lines)
        soup = BeautifulSoup(raw_data, "html.parser")
    else:
        soup = BeautifulSoup(event_data_loc.strip(), "html.parser")
    events_dict = json.loads(soup.text)

    result_dict = {
        "event_id": [],
        "type_id": [],
        "event": [],
        "period_id": [],
        "minutes": [],
        "seconds": [],
        "player_id": [],
        "player_name": [],
        "team_id": [],
        "outcome": [],
        "start_x": [],
        "start_y": [],
        "to_player_id": [],
        "to_player_name": [],
        "end_x": [],
        "end_y": [],
        "td_frame": [],
    }

    for event in events_dict["data"]:
        result_dict["event_id"].append(event["index"])
        result_dict["type_id"].append(event["type"]["id"])
        event_name = event["type"]["name"].lower()
        result_dict["event"].append(event_name)
        result_dict["period_id"].append(event["period"])
        result_dict["minutes"].append(_to_int((event["start"]["time"] // 60)))
        result_dict["seconds"].append(_to_float(event["start"]["time"] % 60))
        result_dict["player_id"].append(_to_int(event["from"]["id"][1:]))
        result_dict["player_name"].append(event["from"]["name"])
        result_dict["team_id"].append(event["team"]["id"])
        result_dict["outcome"].append(np.nan)
        result_dict["start_x"].append(_to_float(event["start"]["x"]))
        result_dict["start_y"].append(_to_float(event["start"]["y"]))
        if event["to"] is not None:
            result_dict["to_player_id"].append(_to_int(event["to"]["id"][1:]))
            result_dict["to_player_name"].append(event["to"]["name"])
        else:
            result_dict["to_player_id"].append(np.nan)
            result_dict["to_player_name"].append(None)
        result_dict["end_x"].append(_to_float(event["end"]["x"]))
        result_dict["end_y"].append(_to_float(event["end"]["y"]))
        result_dict["td_frame"].append(event["start"]["frame"])

    events = pd.DataFrame(result_dict)
    return events



import datetime as dt
from typing import Tuple

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from tqdm import tqdm

from databallpy.load_data.metadata import Metadata
from databallpy.load_data.tracking_data._add_ball_data_to_dict import (
    _add_ball_data_to_dict,
)
from databallpy.load_data.tracking_data._add_periods_to_tracking_data import (
    _add_periods_to_tracking_data,
)
from databallpy.load_data.tracking_data._add_player_tracking_data_to_dict import (
    _add_player_tracking_data_to_dict,
)
from databallpy.load_data.tracking_data._get_matchtime import _get_matchtime
from databallpy.load_data.tracking_data._insert_missing_rows import _insert_missing_rows
from databallpy.load_data.tracking_data._normalize_playing_direction_tracking import (
    _normalize_playing_direction_tracking,
)


def load_tracab_tracking_data(
    tracab_loc: str, metadata_loc: str, verbose: bool = True
) -> Tuple[pd.DataFrame, Metadata]:
    """Function to load tracking data and metadata from the tracab format

    Args:
        tracab_loc (str): location of the tracking_data.dat file
        metadata_loc (str): location of the meta_data.xml file
        verbose (bool): whether to print on progress of loading in the terminal,
        defaults to True

    Returns:
        Tuple[pd.DataFrame, Metadata], the tracking data and metadata class
    """

    tracking_data = _get_tracking_data(tracab_loc, verbose)
    metadata = _get_metadata(metadata_loc)

    tracking_data["period"] = _add_periods_to_tracking_data(
        tracking_data["timestamp"], metadata.periods_frames
    )
    tracking_data["matchtime_td"] = _get_matchtime(
        tracking_data["timestamp"], tracking_data["period"], metadata
    )
    tracking_data = _normalize_playing_direction_tracking(
        tracking_data, metadata.periods_frames
    )

    return tracking_data, metadata


def _get_tracking_data(tracab_loc: str, verbose: bool) -> pd.DataFrame:
    """Function that reads tracking data from .dat file and stores it in a pd.DataFrame

    Args:
        tracab_loc (str): location of the tracking_data.dat file
        verbose (bool): whether to print info in terminal

    Returns:
        pd.DataFrame: contains tracking data
    """

    if verbose:
        print(f"Reading in {tracab_loc}", end="")

    file = open(tracab_loc, "r")
    lines = file.readlines()
    if verbose:
        print(" - Completed")

    file.close()
    size_lines = len(lines)

    data = {
        "timestamp": [np.nan] * size_lines,
        "ball_x": [np.nan] * size_lines,
        "ball_y": [np.nan] * size_lines,
        "ball_z": [np.nan] * size_lines,
        "ball_status": [None] * size_lines,
        "ball_posession": [None] * size_lines,
    }

    if verbose:
        print("Writing lines to dataframe:")
        lines = tqdm(lines)

    for idx, line in enumerate(lines):

        timestamp, players_info, ball_info, _ = line.split(":")
        data["timestamp"][idx] = int(timestamp)

        players = players_info.split(";")[:-1]
        for player in players:
            team_id, _, shirt_num, x, y, _ = player.split(",")
            team_ids = {0: "away", 1: "home"}
            team = team_ids.get(int(team_id))
            if team is None:  # player is unknown or referee
                continue
            data = _add_player_tracking_data_to_dict(team, shirt_num, x, y, data, idx)

        ball_x, ball_y, ball_z, _, posession, status = ball_info.split(";")[0].split(
            ","
        )[:6]
        home_away_map = {"H": "home", "A": "away"}
        posession = home_away_map[posession]
        data = _add_ball_data_to_dict(
            ball_x, ball_y, ball_z, posession, status.lower(), data, idx
        )

    df = pd.DataFrame(data)

    for col in df.columns:
        if "_x" in col or "_y" in col or "_z" in col:
            df[col] = np.round(df[col] / 100, 3)  # change cm to m

    df = _insert_missing_rows(df, "timestamp")

    return df


def _get_metadata(metadata_loc: str) -> Metadata:
    """Function that reads metadata.xml file and stores it in Metadata class

    Args:
        meta_data_loc (str): location of the metadata.xml file

    Returns:
        Metadata: class that contains metadata
    """

    file = open(metadata_loc, "r")
    lines = file.read()
    lines = lines.replace("ï»¿", "")
    soup = BeautifulSoup(lines, "xml")

    match_id = int(soup.find("match")["iId"])
    pitch_size_x = float(soup.find("match")["fPitchXSizeMeters"])
    pitch_size_y = float(soup.find("match")["fPitchYSizeMeters"])
    frame_rate = int(soup.find("match")["iFrameRateFps"])
    datetime_string = soup.find("match")["dtDate"]
    date = pd.to_datetime(datetime_string[:10])

    frames_dict = {
        "period": [],
        "start_frame": [],
        "end_frame": [],
        "start_time_td": [],
        "end_time_td": [],
    }
    for _, period in enumerate(soup.find_all("period")):
        frames_dict["period"].append(int(period["iId"]))
        start_frame = int(period["iStartFrame"])
        end_frame = int(period["iEndFrame"])

        if start_frame != 0:
            frames_dict["start_frame"].append(start_frame)
            frames_dict["end_frame"].append(end_frame)
            frames_dict["start_time_td"].append(
                date + dt.timedelta(milliseconds=int((start_frame / frame_rate) * 1000))
            )
            frames_dict["end_time_td"].append(
                date + dt.timedelta(milliseconds=int((end_frame / frame_rate) * 1000))
            )
        else:
            frames_dict["start_frame"].append(-999)
            frames_dict["end_frame"].append(-999)
            frames_dict["start_time_td"].append(pd.to_datetime("NaT"))
            frames_dict["end_time_td"].append(pd.to_datetime("NaT"))
    df_frames = pd.DataFrame(frames_dict)

    home_team = soup.find("HomeTeam")
    home_team_name = home_team.find("LongName").text
    home_team_id = int(home_team.find("TeamId").text)
    home_players_info = []
    for player in home_team.find_all("Player"):
        player_dict = {}
        for element in player.findChildren():
            player_dict[element.name] = element.text
        home_players_info.append(player_dict)
    df_home_players = _get_players_metadata(home_players_info)

    away_team = soup.find("AwayTeam")
    away_team_name = away_team.find("LongName").text
    away_team_id = int(away_team.find("TeamId").text)

    away_players_info = []
    for player in away_team.find_all("Player"):
        player_dict = {}
        for element in player.findChildren():
            player_dict[element.name] = element.text
        away_players_info.append(player_dict)
    df_away_players = _get_players_metadata(away_players_info)

    metadata = Metadata(
        match_id=match_id,
        pitch_dimensions=[pitch_size_x, pitch_size_y],
        periods_frames=df_frames,
        frame_rate=frame_rate,
        home_team_id=home_team_id,
        home_team_name=home_team_name,
        home_players=df_home_players,
        home_score=np.nan,
        home_formation=None,
        away_team_id=away_team_id,
        away_team_name=away_team_name,
        away_players=df_away_players,
        away_score=np.nan,
        away_formation=None,
    )

    file.close()

    return metadata


def _get_players_metadata(players_info: list) -> pd.DataFrame:
    """Function that creates a df containing info on all players for a team

    Args:
        team (list): contains an information dictionary for each player

    Returns:
        pd.DataFrame: contains all player information for a team
    """

    player_dict = {
        "id": [],
        "full_name": [],
        "shirt_num": [],
        "start_frame": [],
        "end_frame": [],
    }
    for player in players_info:
        player_dict["id"].append(int(player["PlayerId"]))
        full_name = player["FirstName"] + " " + player["LastName"]
        if player["FirstName"] == "":
            full_name = full_name.lstrip()
        player_dict["full_name"].append(full_name)
        player_dict["shirt_num"].append(int(player["JerseyNo"]))
        player_dict["start_frame"].append(int(player["StartFrameCount"]))
        player_dict["end_frame"].append(int(player["EndFrameCount"]))
    df = pd.DataFrame(player_dict)

    return df



import numpy as np


def _add_player_tracking_data_to_dict(
    team: str, shirt_num: str, x: str, y: str, data: dict, idx: int
) -> dict:
    """Function that adds the data of one player to the data dict for one frame

    Args:
        team (str): team side, either "home" or "away"
        shirt_num (str): player's shirt number
        x (float): player's x coordinate at current idx
        y (float): player's y coordinate at current idx
        data (dict): data dictionary to write results to
        idx (int): indicates position in data dictionary

    Returns:
        dict: contains all tracking data
    """

    if f"{team}_{shirt_num}_x" not in data.keys():  # create keys for new player
        data[f"{team}_{shirt_num}_x"] = [np.nan] * len(data["timestamp"])
        data[f"{team}_{shirt_num}_y"] = [np.nan] * len(data["timestamp"])

    data[f"{team}_{shirt_num}_x"][idx] = float(x)
    data[f"{team}_{shirt_num}_y"][idx] = float(y)

    return data



import pandas as pd


def _normalize_playing_direction_events(
    event_data: pd.DataFrame, home_team_id: int, away_team_id: int
) -> pd.DataFrame:
    """Function to normalize the playing direction based on shots, goals and misses so
    that the home team is always represented as playing from left to right and the
    away team from right to left.

    Args:
        event_data (pd.DataFrame): the event data which to normalize
        home_team_id (int): the home team id
        away_team_id (int): the away team id

    Returns:
        pd.DataFrame: normalized event data.
    """
    events = ["shot", "goal", "miss"]
    to_changes_cols = [xy for xy in event_data.columns if "_x" in xy or "_y" in xy]
    for period in event_data["period_id"].unique():
        temp_ed = event_data[event_data["period_id"] == period]

        # home team
        home = temp_ed.loc[
            (temp_ed["team_id"] == home_team_id) & (temp_ed["event"].isin(events))
        ]
        if home["start_x"].mean() < 0:  # home players shoot on the left goal
            event_data.loc[
                temp_ed[temp_ed["team_id"] == home_team_id].index, to_changes_cols
            ] *= -1

        # away team
        away = temp_ed.loc[
            (temp_ed["team_id"] == away_team_id) & (temp_ed["event"].isin(events))
        ]
        if away["start_x"].mean() > 0:  # away players shoot on the right goal
            event_data.loc[
                temp_ed[temp_ed["team_id"] == away_team_id].index, to_changes_cols
            ] *= -1

    return event_data



import datetime as dt
from typing import Tuple

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup

from databallpy.load_data.metadata import Metadata

EVENT_TYPE_IDS = {
    1: "pass",
    2: "offside pass",
    3: "take on",
    4: "foul",
    5: "out",
    6: "corner awarded",
    7: "tackle",
    8: "interception",
    9: "turnover",
    10: "save",
    11: "claim",
    12: "clearance",
    13: "miss",
    14: "post",
    15: "attempt saved",
    16: "goal",
    17: "card",
    18: "player off",
    19: "player on",
    20: "player retired",
    21: "player returns",
    22: "player becomes goalkeeper",
    23: "goalkeeper becomes player",
    24: "condition change",
    25: "official change",
    27: "start delay",
    28: "end delay",
    29: "unknown event 29",
    30: "end",
    31: "unknown event 31",
    32: "start",
    33: "unknown event 33",
    34: "team set up",
    35: "player changed position",
    36: "player changed jersey number",
    37: "collection end",
    38: "temp_goal",
    39: "temp_attempt",
    40: "formation change",
    41: "punch",
    42: "good skill",
    43: "deleted event",
    44: "aerial",
    45: "challenge",
    46: "unknown event 46",
    47: "rescinded card",
    48: "unknown event 46",
    49: "ball recovery",
    50: "dispossessed",
    51: "error",
    52: "keeper pick-up",
    53: "cross not claimed",
    54: "smother",
    55: "offside provoked",
    56: "shield ball opp",
    57: "foul throw in",
    58: "penalty faced",
    59: "keeper sweeper",
    60: "chance missed",
    61: "ball touch",
    62: "unknown event 62",
    63: "temp_save",
    64: "resume",
    65: "contentious referee decision",
    66: "possession data",
    67: "50/50",
    68: "referee drop ball",
    69: "failed to block",
    70: "injury time announcement",
    71: "coach setup",
    72: "caught offside",
    73: "other ball contact",
    74: "blocked pass",
    75: "delayed start",
    76: "early end",
    77: "player off pitch",
}


def load_opta_event_data(
    f7_loc: str, f24_loc: str, pitch_dimensions: list = [106.0, 68.0]
) -> Tuple[pd.DataFrame, Metadata]:
    """This function retrieves the metadata and event data of a specific match. The x
    and y coordinates provided have been scaled to the dimensions of the pitch, with
    (0, 0) being the center. Additionally, the coordinates have been standardized so
    that the home team is represented as playing from left to right for the entire
    match, and the away team is represented as playing from right to left.

    Args:
        f7_loc (str): location of the f7.xml file.
        f24_loc (str): location of the f24.xml file.
        pitch_dimensions (list, optional): the length and width of the pitch in meters

    Returns:
        Tuple[pd.DataFrame, Metadata]: the event data of the match and the metadata
    """

    assert isinstance(f7_loc, str), f"f7_loc should be a string, not a {type(f7_loc)}"
    assert isinstance(
        f24_loc, str
    ), f"f24_loc should be a string, not a {type(f24_loc)}"
    assert f7_loc[-4:] == ".xml", "f7 opta file should by of .xml format"
    assert f24_loc[-4:] == ".xml", "f24 opta file should be of .xml format"

    event_data = _load_event_data(f24_loc)
    metadata = _load_metadata(f7_loc, pitch_dimensions=pitch_dimensions)

    # Add player names to the event data dataframe
    home_players = dict(
        zip(metadata.home_players["id"], metadata.home_players["full_name"])
    )
    away_players = dict(
        zip(metadata.away_players["id"], metadata.away_players["full_name"])
    )

    home_mask = (event_data["team_id"] == metadata.home_team_id) & ~pd.isnull(
        event_data["player_id"]
    )
    away_mask = (event_data["team_id"] == metadata.away_team_id) & ~pd.isnull(
        event_data["player_id"]
    )

    event_data.loc[home_mask, "player_name"] = event_data.loc[
        home_mask, "player_id"
    ].map(home_players)
    event_data.loc[away_mask, "player_name"] = event_data.loc[
        away_mask, "player_id"
    ].map(away_players)

    # Rescale the x and y coordinates relative to the pitch dimensions
    # The original dimension of the x and y coordinates range from 0 to 100
    event_data.loc[:, ["start_x"]] = (
        event_data.loc[:, ["start_x"]] / 100 * pitch_dimensions[0]
    ) - (pitch_dimensions[0] / 2.0)
    event_data.loc[:, ["start_y"]] = (
        event_data.loc[:, ["start_y"]] / 100 * pitch_dimensions[1]
    ) - (pitch_dimensions[1] / 2.0)

    # Change direction of play of the away team so it is represented from right to left
    event_data.loc[
        event_data["team_id"] == metadata.away_team_id, ["start_x", "start_y"]
    ] *= -1

    return event_data, metadata


def _load_metadata(f7_loc: str, pitch_dimensions: list) -> Metadata:
    """Function to load metadata from the f7.xml opta file

    Args:
        f7_loc (str): location of the f7.xml opta file
        pitch_dimensions (list): the length and width of the pitch in meters

    Returns:
        MetaData: all metadata information of the current match
    """
    file = open(f7_loc, "r")
    lines = file.read()
    soup = BeautifulSoup(lines, "xml")

    # Obtain match id
    match_id = int(soup.find("SoccerDocument").attrs["uID"][1:])

    # Obtain match start and end of period datetime
    periods = {
        "period": [1, 2, 3, 4, 5],
        "start_datetime_opta": [],
        "end_datetime_opta": [],
    }
    start_period_1 = soup.find("Stat", attrs={"Type": "first_half_start"})
    end_period_1 = soup.find("Stat", attrs={"Type": "first_half_stop"})
    start_period_2 = soup.find("Stat", attrs={"Type": "second_half_start"})
    end_period_2 = soup.find("Stat", attrs={"Type": "second_half_stop"})
    for start, end in zip(
        [start_period_1, start_period_2], [end_period_1, end_period_2]
    ):
        # Add one hour to go from utc to Dutch time
        periods["start_datetime_opta"].append(
            pd.to_datetime(start.contents[0]).tz_localize(None) + dt.timedelta(hours=1)
        )
        periods["end_datetime_opta"].append(
            pd.to_datetime(end.contents[0]).tz_localize(None) + dt.timedelta(hours=1)
        )
    for _ in range(3):
        periods["start_datetime_opta"].append(pd.to_datetime("NaT"))
        periods["end_datetime_opta"].append(pd.to_datetime("NaT"))

    # Opta has a TeamData and Team attribute in the f7 file
    team_datas = soup.find_all("TeamData")
    teams = soup.find_all("Team")
    teams_info = {}
    teams_player_info = {}
    for team_data, team in zip(team_datas, teams):

        # Team information
        team_name = team.findChildren("Name")[0].contents[0]
        team_info = {}
        team_info["team_name"] = team_name
        team_info["side"] = team_data["Side"].lower()
        team_info["formation"] = team_data["Formation"]
        team_info["score"] = int(team_data["Score"])
        team_info["team_id"] = int(team_data["TeamRef"][1:])
        teams_info[team_info["side"]] = team_info

        # Player information
        players_data = [
            player.attrs for player in team_data.findChildren("MatchPlayer")
        ]
        players_names = {}
        for player in team.findChildren("Player"):
            player_id = int(player.attrs["uID"][1:])
            first_name = player.contents[1].contents[1].text

            if "Last" in str(player.contents[1].contents[3]):
                last_name_idx = 3
            else:
                last_name_idx = 5
            last_name = player.contents[1].contents[last_name_idx].contents[0]
            if first_name:
                players_names[str(player_id)] = f"{first_name} {last_name}"
            else:
                players_names[str(player_id)] = last_name

        player_info = _get_player_info(players_data, players_names)
        teams_player_info[team_info["side"]] = player_info

    file.close()

    metadata = Metadata(
        match_id=match_id,
        pitch_dimensions=pitch_dimensions,
        periods_frames=pd.DataFrame(periods),
        frame_rate=np.nan,
        home_team_id=teams_info["home"]["team_id"],
        home_team_name=str(teams_info["home"]["team_name"]),
        home_players=teams_player_info["home"],
        home_score=teams_info["home"]["score"],
        home_formation=teams_info["home"]["formation"],
        away_team_id=teams_info["away"]["team_id"],
        away_team_name=str(teams_info["away"]["team_name"]),
        away_players=teams_player_info["away"],
        away_score=teams_info["away"]["score"],
        away_formation=teams_info["away"]["formation"],
    )
    return metadata


def _get_player_info(players_data: list, players_names: dict) -> pd.DataFrame:
    """Function to loop over all players and save data in a pd.DataFrame

    Args:
        players_data (list): for every player a dictionary with info about the player
        except the player name
        players_names (dict): dictionary with player id as key and the player name as
        value

    Returns:
        pd.DataFrame: all information of the players
    """
    result_dict = {
        "id": [],
        "full_name": [],
        "formation_place": [],
        "position": [],
        "starter": [],
        "shirt_num": [],
    }

    for player in players_data:
        player_id = int(player["PlayerRef"][1:])
        result_dict["id"].append(player_id)
        result_dict["full_name"].append(players_names[str(player_id)])
        result_dict["formation_place"].append(int(player["Formation_Place"]))
        position = (
            player["Position"]
            if player["Position"] != "Substitute"
            else player["SubPosition"]
        )
        result_dict["position"].append(position.lower())
        result_dict["starter"].append(player["Status"] == "Start")
        result_dict["shirt_num"].append(int(player["ShirtNumber"]))

    return pd.DataFrame(result_dict)


def _load_event_data(f24_loc: str) -> pd.DataFrame:
    """Function to load the f27 .xml, the events of the match.
    Note: this function does ignore qualifiers for now

    Args:
        f24_loc (str): location of the f24.xml file

    Returns:
        pd.DataFrame: all events of the match in a pd dataframe
    """

    file = open(f24_loc, "r")
    lines = file.read()
    soup = BeautifulSoup(lines, "xml")

    result_dict = {
        "event_id": [],
        "type_id": [],
        "event": [],
        "period_id": [],
        "minutes": [],
        "seconds": [],
        "player_id": [],
        "team_id": [],
        "outcome": [],
        "start_x": [],
        "start_y": [],
        "datetime": [],
    }

    events = soup.find_all("Event")
    for event in events:
        result_dict["event_id"].append(int(event.attrs["id"]))
        event_type_id = int(event.attrs["type_id"])
        result_dict["type_id"].append(event_type_id)

        if event_type_id in EVENT_TYPE_IDS.keys():
            event_name = EVENT_TYPE_IDS[event_type_id].lower()
        else:
            # Unknown event
            event_name = None

        result_dict["event"].append(event_name)
        result_dict["period_id"].append(int(event.attrs["period_id"]))
        result_dict["minutes"].append(int(event.attrs["min"]))
        result_dict["seconds"].append(int(event.attrs["sec"]))

        if "player_id" in event.attrs.keys():
            result_dict["player_id"].append(int(event.attrs["player_id"]))
        else:
            result_dict["player_id"].append(np.nan)

        result_dict["team_id"].append(int(event.attrs["team_id"]))
        result_dict["outcome"].append(int(event.attrs["outcome"]))
        result_dict["start_x"].append(float(event.attrs["x"]))
        result_dict["start_y"].append(float(event.attrs["y"]))
        result_dict["datetime"].append(
            pd.to_datetime(event.attrs["timestamp"]) + dt.timedelta(hours=1)
        )

    file.close()
    return pd.DataFrame(result_dict)



import datetime as dt
from dataclasses import dataclass
from typing import List

import numpy as np
import pandas as pd
from tqdm import tqdm

from databallpy import DataBallPyError
from databallpy.load_data.event_data.metrica_event_data import (
    load_metrica_event_data,
    load_metrica_open_event_data,
)
from databallpy.load_data.event_data.opta import load_opta_event_data
from databallpy.load_data.tracking_data.metrica_tracking_data import (
    load_metrica_open_tracking_data,
    load_metrica_tracking_data,
)
from databallpy.load_data.tracking_data.tracab import load_tracab_tracking_data


@dataclass
class Match:
    """This is the match class. It contains all information of the match and has
    some simple functions to easily obtain information about the match.

    Args:
        tracking_data (pd.DataFrame): Tracking data of the match.
        tracking_data_provider (str): Provider of the tracking data.
        event_data (pd.DataFrame): Event data of the match.
        event_data_provider (str): Provider of the event data.
        pitch_dimensions (Tuple): The size of the pitch in meters in x and y direction.
        periods (pd.DataFrame): The start and end idicators of all periods.
        frame_rate (int): The frequency of the tracking data.
        home_team_id (int): The id of the home team.
        home_team_name (str): The name of the home team.
        home_players (pd.DataFrame): Information about the home players.
        home_score (int): Number of goals scored over the match by the home team.
        home_formation (str): Indication of the formation of the home team.
        away_team_id (int): The id of the away team.
        away_team_name (str): The name of the away team.
        away_players (pd.DataFrame): Information about the away players.
        away_score (int): Number of goals scored over the match by the away team.
        away_formation (str): Indication of the formation of the away team.
        name (str): The home and away team name and score (e.g "nameH 3 - 1 nameA").
        home_players_column_ids (list): All column ids of the tracking data that refer
                                        to information about the home team players.
        away_players_column_ids (list): All column ids of the tracking data that refer
                                        to information about the away team players.

    Funcs
        player_column_id_to_full_name: Simple function to get the full name of a player
                                       from the column id
    """

    tracking_data: pd.DataFrame
    tracking_data_provider: str
    event_data: pd.DataFrame
    event_data_provider: str
    pitch_dimensions: List[float]
    periods: pd.DataFrame
    frame_rate: int
    home_team_id: int
    home_team_name: str
    home_players: pd.DataFrame
    home_score: int
    home_formation: str
    away_team_id: int
    away_team_name: str
    away_players: pd.DataFrame
    away_score: int
    away_formation: str

    def __post_init__(self):
        # tracking_data
        if not isinstance(self.tracking_data, pd.DataFrame):
            raise TypeError(
                f"tracking data should be a pandas df, not a {type(self.tracking_data)}"
            )
        for col in ["timestamp", "ball_x", "ball_y"]:
            if col not in self.tracking_data.columns.to_list():
                raise ValueError(
                    f"No {col} in tracking_data columns, this is manditory!"
                )

        # tracking_data_provider
        if not isinstance(self.tracking_data_provider, str):
            raise TypeError(
                f"tracking data provider should be a string, not a \
                    {type(self.tracking_data_provider)}"
            )

        # event_data
        if not isinstance(self.event_data, pd.DataFrame):
            raise TypeError(
                f"event data should be a pandas df, not a {type(self.event_data)}"
            )
        for col in [
            "event_id",
            "event",
            "period_id",
            "team_id",
            "player_id",
            "start_x",
            "start_y",
            "datetime",
        ]:
            if col not in self.event_data.columns.to_list():
                raise ValueError(f"{col} not in event data columns, this is manditory!")

        # event_data_provider
        if not isinstance(self.event_data_provider, str):
            raise TypeError(
                f"event data provider should be a string, not a \
                    {type(self.event_data_provider)}"
            )

        # pitch_dimensions
        if not isinstance(self.pitch_dimensions, list):
            raise TypeError(
                f"pitch_dimensions ({self.pitch_dimensions}) should be a \
                    list, not a {type(self.pitch_dimensions)}"
            )
        if not len(self.pitch_dimensions) == 2:
            raise ValueError(
                f"pitch_dimensions should contain, two values: a length and a width \
                    of the pitch, current input is {self.pitch_dimensions}"
            )
        if not all([isinstance(x, float) for x in self.pitch_dimensions]):
            raise TypeError(
                f"Both values in pitch dimensions should by floats, current inputs \
                    {[type(x) for x in self.pitch_dimensions]}"
            )

        # periods
        if not isinstance(self.periods, pd.DataFrame):
            raise TypeError(
                f"periods_frames should be a pandas dataframe, not a \
                    {type(self.periods)}"
            )
        if "period" not in self.periods.columns:
            raise ValueError("'period' should be one of the columns in period_frames")
        if any(
            [
                x not in self.periods["period"].value_counts().index
                for x in [1, 2, 3, 4, 5]
            ]
        ) or not all(self.periods["period"].value_counts() == 1):

            res = self.periods["period"]
            raise ValueError(
                f"'period' column in period_frames should contain only the values \
                    [1, 2, 3, 4, 5]. Now it's {res}"
            )

        # frame_rate
        if not pd.isnull(self.frame_rate):
            if not isinstance(self.frame_rate, int):
                raise TypeError(
                    f"frame_rate should be an integer, not a {type(self.frame_rate)}"
                )
            if self.frame_rate < 1:
                raise ValueError(
                    f"frame_rate should be positive, not {self.frame_rate}"
                )

        # team id's
        for team, team_id in zip(
            ["home", "away"], [self.home_team_id, self.away_team_id]
        ):
            if not isinstance(team_id, int) and not isinstance(team_id, str):
                raise TypeError(
                    f"{team} team id should be an integer or string, not a \
                        {type(team_id)}"
                )

        # team names
        for team, name in zip(
            ["home", "away"], [self.home_team_name, self.away_team_name]
        ):
            if not isinstance(name, str):
                raise TypeError(
                    f"{team} team name should be a string, not a {type(name)}"
                )

        # team scores
        for team, score in zip(["home", "away"], [self.home_score, self.away_score]):
            if not pd.isnull(score):
                if not isinstance(score, int):
                    raise TypeError(
                        f"{team} team score should be an integer, not a {type(score)}"
                    )
                if score < 0:
                    raise ValueError(f"{team} team score should positive, not {score}")

        # team formations
        for team, form in zip(
            ["home", "away"], [self.home_formation, self.away_formation]
        ):
            if not isinstance(form, str):
                raise TypeError(
                    f"{team} team formation should be a string, not a {type(form)}"
                )
            if len(form) > 4:
                raise ValueError(
                    f"{team} team formation should be of length 4 or smaller \
                        ('1433'), not {len(form)}"
                )

        # team players
        for team, players in zip(
            ["home", "away"], [self.home_players, self.away_players]
        ):
            if not isinstance(players, pd.DataFrame):
                raise TypeError(
                    f"{team} team players should be a pandas dataframe, not a \
                        {type(players)}"
                )
            for col in ["id", "full_name", "shirt_num"]:
                if col not in players.columns:
                    raise ValueError(
                        f"{team} team players should contain at least the column \
                            ['id', 'full_name', 'shirt_num'], {col} is missing."
                    )

        # check for pitch axis
        if (
            not abs(
                self.tracking_data["ball_x"].min() + self.tracking_data["ball_x"].max()
            )
            < 5.0
        ):
            max_x = self.tracking_data["ball_x"].max()
            min_x = self.tracking_data["ball_x"].min()
            raise DataBallPyError(
                f"The middle point of the pitch should be (0, 0),\
                                now the min x = {min_x} and the max x = {max_x}"
            )

        if (
            not abs(
                self.tracking_data["ball_y"].min() + self.tracking_data["ball_y"].max()
            )
            < 5.0
        ):
            max_y = self.tracking_data["ball_y"].max()
            min_y = self.tracking_data["ball_y"].min()
            raise DataBallPyError(
                f"The middle point of the pitch should be (0, 0),\
                                now th min y = {min_y} and the max y = {max_y}"
            )

        # check for direction of play
        for _, period_row in self.periods.iterrows():
            frame = period_row["start_frame"]
            if (
                len(self.tracking_data[self.tracking_data["timestamp"] == frame].index)
                == 0
            ):
                continue
            idx = self.tracking_data[self.tracking_data["timestamp"] == frame].index[0]
            period = period_row["period"]
            home_x = [x for x in self.home_players_column_ids if "_x" in x]
            away_x = [x for x in self.away_players_column_ids if "_x" in x]
            if self.tracking_data.loc[idx, home_x].mean() > 0:
                centroid_x = self.tracking_data.loc[idx, home_x].mean()
                raise DataBallPyError(
                    f"The home team should be represented as playing\
from left to right the whole match. At the start of period {period} the x centroid of \
the home team is {centroid_x}."
                )

            if self.tracking_data.loc[idx, away_x].mean() < 0:
                centroid_x = self.tracking_data.loc[idx, away_x].mean()
                raise DataBallPyError(
                    f"The away team should be represented as playing\
from right to left the whole match. At the start  of period {period} the x centroid of \
the away team is {centroid_x}."
                )

    @property
    def name(self) -> str:
        home_text = f"{self.home_team_name} {self.home_score}"
        away_text = f"{self.away_score} {self.away_team_name}"
        return f"{home_text} - {away_text}"

    @property
    def home_players_column_ids(self) -> List[str]:
        return [id for id in self.tracking_data.columns if "home" in id]

    @property
    def away_players_column_ids(self) -> List[str]:
        return [id for id in self.tracking_data.columns if "away" in id]

    def player_column_id_to_full_name(self, column_id: str) -> str:
        """Simple function to get the full name of a player from the column id

        Args:
            column_id (str): the column id of a player, for instance "home_1"

        Returns:
            str: full name of the player
        """
        shirt_num = int(column_id.split("_")[1])
        if column_id[:4] == "home":
            return self.home_players.loc[
                self.home_players["shirt_num"] == shirt_num, "full_name"
            ].iloc[0]
        else:
            return self.away_players.loc[
                self.away_players["shirt_num"] == shirt_num, "full_name"
            ].iloc[0]

    def player_id_to_column_id(self, player_id: int) -> str:
        """Simple function to get the column id based on player id

        Args:
            player_id (int): id of the player

        Returns:
            str: column id of the player, for instance "home_1"
        """
        if (self.home_players["id"].eq(player_id)).any():
            num = self.home_players[self.home_players["id"] == player_id][
                "shirt_num"
            ].iloc[0]
            return f"home_{num}"
        elif (self.away_players["id"].eq(player_id)).any():
            num = self.away_players[self.away_players["id"] == player_id][
                "shirt_num"
            ].iloc[0]
            return f"away_{num}"
        else:
            raise ValueError(f"{player_id} is not in either one of the teams")

    def synchronise_tracking_and_event_data(self, n_batches_per_half: int = 9):
        """Function that synchronises tracking and event data using Needleman-Wunsch
           algorithmn. Based on: https://kwiatkowski.io/sync.soccer

        Args:
            n_batches_per_half (int): the number of batches that are created per half.
            A higher number of batches reduces the time the code takes to load, but
            reduces the accuracy for events close to the splits.
            Default = 9

        Currently works for the following events:
            'pass', 'aerial', 'interception', 'ball recovery', 'dispossessed', 'tackle',
            'take on', 'clearance', 'blocked pass', 'offside pass', 'attempt saved',
            'save', 'foul', 'miss', 'challenge', 'goal'

        """
        events_to_sync = [
            "pass",
            "aerial",
            "interception",
            "ball recovery",
            "dispossessed",
            "tackle",
            "take on",
            "clearance",
            "blocked pass",
            "offside pass",
            "attempt saved",
            "save",
            "foul",
            "miss",
            "challenge",
            "goal",
        ]

        tracking_data = self.tracking_data
        event_data = self.event_data

        start_datetime_period = {}
        start_frame_period = {}
        for _, row in self.periods.iterrows():
            start_datetime_period[row["period"]] = row["start_time_td"]
            start_frame_period[row["period"]] = row["start_frame"]

        tracking_data["datetime"] = [
            start_datetime_period[int(p)]
            + dt.timedelta(
                milliseconds=int((x - start_frame_period[p]) / self.frame_rate * 1000)
            )
            if p > 0
            else pd.to_datetime("NaT")
            for x, p in zip(tracking_data["timestamp"], tracking_data["period"])
        ]

        tracking_data["event"] = np.nan
        tracking_data["event_id"] = np.nan
        event_data["tracking_frame"] = np.nan

        mask_events_to_sync = event_data["event"].isin(events_to_sync)
        event_data = event_data[mask_events_to_sync]

        periods_played = self.periods[self.periods["start_frame"] > 0]["period"].values

        for p in periods_played:
            # create batches to loop over
            start_batch_frame = self.periods.loc[
                self.periods["period"] == p, "start_frame"
            ].iloc[0]
            start_batch_datetime = start_datetime_period[p] + dt.timedelta(
                milliseconds=int(
                    (start_batch_frame - start_frame_period[p]) / self.frame_rate * 1000
                )
            )
            delta = (
                self.periods.loc[self.periods["period"] == p, "end_frame"].iloc[0]
                - start_batch_frame
            )
            end_batches_frames = np.floor(
                np.arange(
                    delta / n_batches_per_half,
                    delta + delta / n_batches_per_half,
                    delta / n_batches_per_half,
                )
                + start_batch_frame
            )
            end_batches_datetime = [
                start_datetime_period[int(p)]
                + dt.timedelta(
                    milliseconds=int(
                        (x - start_frame_period[int(p)]) / self.frame_rate * 1000
                    )
                )
                for x in end_batches_frames
            ]

            tracking_data_period = tracking_data[tracking_data["period"] == p]
            event_data_period = event_data[event_data["period_id"] == p].copy()
            start_events = ["pass", "miss", "goal"]
            datetime_first_event = event_data_period[
                event_data_period["event"].isin(start_events)
            ].iloc[0]["datetime"]
            datetime_first_tracking_frame = tracking_data_period.iloc[0]["datetime"]
            diff_datetime = datetime_first_event - datetime_first_tracking_frame
            event_data_period["datetime"] -= diff_datetime

            print(f"Syncing period {p}...")
            for end_batch_frame, end_batch_datetime in tqdm(
                zip(end_batches_frames, end_batches_datetime),
                total=len(end_batches_frames),
            ):

                tracking_batch = tracking_data_period[
                    (tracking_data_period["timestamp"] <= end_batch_frame)
                    & (tracking_data_period["timestamp"] >= start_batch_frame)
                ].reset_index(drop=False)
                event_batch = event_data_period[
                    (event_data_period["datetime"] >= start_batch_datetime)
                    & (event_data_period["datetime"] <= end_batch_datetime)
                ].reset_index(drop=False)

                sim_mat = _create_sim_mat(tracking_batch, event_batch, self)
                event_frame_dict = _needleman_wunsch(sim_mat)

                for event, frame in event_frame_dict.items():
                    event_id = int(event_batch.loc[event, "event_id"])
                    event_type = event_batch.loc[event, "event"]
                    event_index = int(event_batch.loc[event, "index"])
                    tracking_frame = int(tracking_batch.loc[frame, "index"])
                    tracking_data.loc[tracking_frame, "event"] = event_type
                    tracking_data.loc[tracking_frame, "event_id"] = event_id
                    event_data.loc[event_index, "tracking_frame"] = tracking_frame

                start_batch_frame = tracking_data_period.loc[
                    tracking_frame, "timestamp"
                ]
                start_batch_datetime = event_data_period[
                    event_data_period["event_id"] == event_id
                ]["datetime"].iloc[0]
        tracking_data.drop("datetime", axis=1, inplace=True)
        self.tracking_data = tracking_data
        self.event_data = event_data

        return

    def __eq__(self, other):
        if isinstance(other, Match):
            res = [
                self.tracking_data.equals(other.tracking_data),
                self.tracking_data_provider == other.tracking_data_provider,
                self.event_data.equals(other.event_data),
                self.pitch_dimensions == other.pitch_dimensions,
                self.periods.equals(other.periods),
                self.frame_rate == other.frame_rate,
                self.home_team_id == other.home_team_id,
                self.home_team_name == other.home_team_name,
                self.home_formation == other.home_formation,
                self.home_players.equals(other.home_players),
                self.home_score == other.home_score,
                self.away_team_id == other.away_team_id,
                self.away_team_name == other.away_team_name,
                self.away_formation == other.away_formation,
                self.away_players.equals(other.away_players),
                self.away_score == other.away_score,
            ]

            return all(res)
        else:
            return False


def get_match(
    *,
    tracking_data_loc: str,
    tracking_metadata_loc: str,
    event_data_loc: str,
    event_metadata_loc: str,
    tracking_data_provider: str,
    event_data_provider: str,
) -> Match:
    """Function to get all information of a match given its datasources

    Args:
        tracking_data_loc (str): location of the tracking data
        tracking_metadata_loc (str): location of the metadata of the tracking data
        event_data_loc (str): location of the event data
        event_metadata_loc (str): location of the metadata of the event data
        tracking_data_provider (str): provider of the tracking data
        event_data_provider (str): provider of the event data
    Returns:
        (Match): a Match object with all information of the match.
    """

    assert tracking_data_provider in [
        "tracab",
        "metrica",
    ], f"We do not support '{tracking_data_provider}' as tracking data provider yet, "
    "please open an issue in our Github repository."
    assert event_data_provider in [
        "opta",
        "metrica",
    ], f"We do not supper '{event_data_provider}' as event data provider yet, "
    "please open an issue in our Github repository."

    # Get event data and event metadata
    if event_data_provider == "opta":
        event_data, event_metadata = load_opta_event_data(
            f7_loc=event_metadata_loc, f24_loc=event_data_loc
        )
    elif event_data_provider == "metrica":
        event_data, event_metadata = load_metrica_event_data(
            event_data_loc=event_data_loc, metadata_loc=event_metadata_loc
        )

    # Get tracking data and tracking metadata
    if tracking_data_provider == "tracab":
        tracking_data, tracking_metadata = load_tracab_tracking_data(
            tracking_data_loc, tracking_metadata_loc
        )
    elif tracking_data_provider == "metrica":
        tracking_data, tracking_metadata = load_metrica_tracking_data(
            tracking_data_loc=tracking_data_loc, metadata_loc=tracking_metadata_loc
        )

    # Check if the event data is scaled the right way
    if not tracking_metadata.pitch_dimensions == event_metadata.pitch_dimensions:
        x_correction = (
            tracking_metadata.pitch_dimensions[0] / event_metadata.pitch_dimensions[0]
        )
        y_correction = (
            tracking_metadata.pitch_dimensions[1] / event_metadata.pitch_dimensions[1]
        )
        event_data["start_x"] *= x_correction
        event_data["start_y"] *= y_correction

    if tracking_data_provider == event_data_provider == "metrica":
        merged_periods = tracking_metadata.periods_frames
        home_players = tracking_metadata.home_players
        away_players = tracking_metadata.away_players
    else:
        # Merge periods
        merged_periods = pd.concat(
            (
                tracking_metadata.periods_frames,
                event_metadata.periods_frames.drop("period", axis=1),
            ),
            axis=1,
        )

        # Merged player info
        home_players = tracking_metadata.home_players.merge(
            event_metadata.home_players[
                ["id", "formation_place", "position", "starter"]
            ],
            on="id",
        )
        away_players = tracking_metadata.away_players.merge(
            event_metadata.away_players[
                ["id", "formation_place", "position", "starter"]
            ],
            on="id",
        )

    match = Match(
        tracking_data=tracking_data,
        tracking_data_provider=tracking_data_provider,
        event_data=event_data,
        event_data_provider=event_data_provider,
        pitch_dimensions=tracking_metadata.pitch_dimensions,
        periods=merged_periods,
        frame_rate=tracking_metadata.frame_rate,
        home_team_id=event_metadata.home_team_id,
        home_formation=event_metadata.home_formation,
        home_score=event_metadata.home_score,
        home_team_name=event_metadata.home_team_name,
        home_players=home_players,
        away_team_id=event_metadata.away_team_id,
        away_formation=event_metadata.away_formation,
        away_score=event_metadata.away_score,
        away_team_name=event_metadata.away_team_name,
        away_players=away_players,
    )

    return match


def _create_sim_mat(
    tracking_batch: pd.DataFrame, event_batch: pd.DataFrame, match: Match
) -> np.ndarray:
    """Function that creates similarity matrix between every frame and event in batch

    Args:
        tracking_batch (pd.DataFrame): batch of tracking data
        event_batch (pd.DataFrame): batch of event data
        match (Match): Match class containing full_name_to_player_column_id function

    Returns:
        np.ndarray: array containing similarity scores between every frame and events,
                    size is #frames, #events
    """
    sim_mat = np.zeros((len(tracking_batch), len(event_batch)))
    tracking_batch["datetime"] = tracking_batch["datetime"]

    for i, event in event_batch.iterrows():
        time_diff = (tracking_batch["datetime"] - event["datetime"]) / dt.timedelta(
            seconds=1
        )
        ball_loc_diff = np.hypot(
            tracking_batch["ball_x"] - event["start_x"],
            tracking_batch["ball_y"] - event["start_y"],
        )

        if not np.isnan(event["player_id"]):
            column_id_player = match.player_id_to_column_id(
                player_id=event["player_id"]
            )
            player_ball_diff = np.hypot(
                tracking_batch["ball_x"] - tracking_batch[f"{column_id_player}_x"],
                tracking_batch["ball_y"] - tracking_batch[f"{column_id_player}_y"],
            )
        else:
            player_ball_diff = 0
        # similarity function from: https://kwiatkowski.io/sync.soccer
        sim_mat[:, i] = np.abs(time_diff) + ball_loc_diff / 5 + player_ball_diff

    sim_mat[np.isnan(sim_mat)] = np.nanmax(
        sim_mat
    )  # replace nan values with highest value
    den = np.nanmax(np.nanmin(sim_mat, axis=1))  # scale similarity scores
    sim_mat = np.exp(-sim_mat / den)

    return sim_mat


def _needleman_wunsch(
    sim_mat: np.ndarray, gap_event: int = -1, gap_frame: int = 1
) -> dict:
    """
    Function that calculates the optimal alignment between events and frames
    given similarity scores between all frames and events
    Based on: https://gist.github.com/slowkow/06c6dba9180d013dfd82bec217d22eb5

    Args:
        sim_mat (np.ndarray): matrix with similarity between every frame and event
        gap_event (int): penalty for leaving an event unassigned to a frame
                         (not allowed), defaults to -1
        gap_frame (int): penalty for leaving a frame unassigned to a penalty
                         (very common), defaults to 1

    Returns:
       event_frame_dict (dict): dictionary with events as keys and frames as values
    """
    n_frames, n_events = np.shape(sim_mat)

    F = np.zeros((n_frames + 1, n_events + 1))
    F[:, 0] = np.linspace(0, n_frames * gap_frame, n_frames + 1)
    F[0, :] = np.linspace(0, n_events * gap_event, n_events + 1)

    # Pointer matrix
    P = np.zeros((n_frames + 1, n_events + 1))
    P[:, 0] = 3
    P[0, :] = 4

    t = np.zeros(3)
    for i in range(n_frames):
        for j in range(n_events):
            t[0] = F[i, j] + sim_mat[i, j]
            t[1] = F[i, j + 1] + gap_frame  # top + gap frame
            t[2] = F[i + 1, j] + gap_event  # left + gap event
            tmax = np.max(t)
            F[i + 1, j + 1] = tmax

            if t[0] == tmax:  # match
                P[i + 1, j + 1] += 2
            if t[1] == tmax:  # frame unassigned
                P[i + 1, j + 1] += 3
            if t[2] == tmax:  # event unassigned
                P[i + 1, j + 1] += 4

    # Trace through an optimal alignment.
    i = n_frames
    j = n_events
    frames = []
    events = []
    while i > 0 or j > 0:
        if P[i, j] in [2, 5, 6, 9]:  # 2 was added, match
            frames.append(i)
            events.append(j)
            i -= 1
            j -= 1
        elif P[i, j] in [3, 5, 7, 9]:  # 3 was added, frame unassigned
            frames.append(i)
            events.append(0)
            i -= 1
        elif P[i, j] in [4, 6, 7, 9]:  # 4 was added, event unassigned
            raise ValueError(
                "An event was left unassigned, check your gap penalty values"
            )

    frames = frames[::-1]
    events = events[::-1]

    idx_events = [idx for idx, i in enumerate(events) if i > 0]
    event_frame_dict = {}
    for i in idx_events:
        event_frame_dict[events[i] - 1] = frames[i] - 1

    return event_frame_dict


def get_open_match(provider: str = "metrica") -> Match:
    """Function to load a match object from an open datasource

    Args:
        provider (str, optional): What provider to get the open data from.
        Defaults to "metrica".

    Returns:
        Match: All information about the match
    """
    assert provider in ["metrica"]

    if provider == "metrica":
        tracking_data, metadata = load_metrica_open_tracking_data()
        event_data, _ = load_metrica_open_event_data()

    match = Match(
        tracking_data=tracking_data,
        tracking_data_provider=provider,
        event_data=event_data,
        event_data_provider=provider,
        pitch_dimensions=metadata.pitch_dimensions,
        periods=metadata.periods_frames,
        frame_rate=metadata.frame_rate,
        home_team_id=metadata.home_team_id,
        home_formation=metadata.home_formation,
        home_score=metadata.home_score,
        home_team_name=metadata.home_team_name,
        home_players=metadata.home_players,
        away_team_id=metadata.away_team_id,
        away_formation=metadata.away_formation,
        away_score=metadata.away_score,
        away_team_name=metadata.away_team_name,
        away_players=metadata.away_players,
    )
    return match



import io
import os
from typing import Tuple, Union

import numpy as np
import pandas as pd
import requests
from tqdm import tqdm

from databallpy.load_data.metadata import Metadata
from databallpy.load_data.metrica_metadata import (
    _get_metadata,
    _get_td_channels,
    _update_metadata,
)
from databallpy.load_data.tracking_data._add_ball_data_to_dict import (
    _add_ball_data_to_dict,
)
from databallpy.load_data.tracking_data._add_periods_to_tracking_data import (
    _add_periods_to_tracking_data,
)
from databallpy.load_data.tracking_data._add_player_tracking_data_to_dict import (
    _add_player_tracking_data_to_dict,
)
from databallpy.load_data.tracking_data._get_matchtime import _get_matchtime
from databallpy.load_data.tracking_data._insert_missing_rows import _insert_missing_rows
from databallpy.load_data.tracking_data._normalize_playing_direction_tracking import (
    _normalize_playing_direction_tracking,
)
from databallpy.utils import _to_int


def load_metrica_tracking_data(
    tracking_data_loc: str, metadata_loc: str, verbose: bool = True
) -> Tuple[pd.DataFrame, Metadata]:
    """Function to load metrica tracking data.

    Args:
        tracking_data_loc (str): location of the tracking data .txt file
        metadata_loc (str): location of the metadata .xml file
        verbose (bool, optional): whether to print information about the progress
        in the terminall. Defaults to True.

    Raises:
        TypeError: if tracking_data_loc is not a string or io.StringIO

    Returns:
        Tuple[pd.DataFrame, Metadata]: tracking and metadata of the match
    """

    if isinstance(tracking_data_loc, str):
        assert os.path.exists(tracking_data_loc)
        assert os.path.exists(metadata_loc)
    elif isinstance(tracking_data_loc, io.StringIO):
        pass
    else:
        raise TypeError(
            "tracking_data_loc must be either a str or a StringIO object,"
            f" not a {type(tracking_data_loc)}"
        )

    metadata = _get_metadata(metadata_loc)
    td_channels = _get_td_channels(metadata_loc, metadata)
    metadata = _update_metadata(td_channels, metadata)
    tracking_data = _get_tracking_data(
        tracking_data_loc, td_channels, metadata.pitch_dimensions, verbose=verbose
    )

    tracking_data = _normalize_playing_direction_tracking(
        tracking_data, metadata.periods_frames
    )
    tracking_data["period"] = _add_periods_to_tracking_data(
        tracking_data["timestamp"], metadata.periods_frames
    )
    tracking_data["matchtime_td"] = _get_matchtime(
        tracking_data["timestamp"], tracking_data["period"], metadata
    )

    return tracking_data, metadata


def load_metrica_open_tracking_data() -> Tuple[pd.DataFrame, Metadata]:
    """Function to load open dataset of metrica

    Returns:
        Tuple[pd.DataFrame, Metadata]: tracking and metadata of the match
    """
    td_data_link = "https://raw.githubusercontent.com/metrica-sports/sample-data\
        /master/data/Sample_Game_3/Sample_Game_3_tracking.txt"
    td_metadata_link = "https://raw.githubusercontent.com/metrica-sports/sample-data\
        /master/data/Sample_Game_3/Sample_Game_3_metadata.xml"

    print("Downloading Metrica open tracking data...", end="")
    td_data = io.StringIO(requests.get(td_data_link).text)
    td_metadata = requests.get(td_metadata_link).text
    print(" Done!")
    return load_metrica_tracking_data(
        tracking_data_loc=td_data, metadata_loc=td_metadata
    )


def _get_tracking_data(
    tracking_data_loc: Union[str, io.StringIO],
    channels: dict,
    pitch_dimensions: list,
    verbose: bool = True,
) -> pd.DataFrame:
    """Function to load the tracking data of metrica.

    Args:
        tracking_data_loc (Union[str, io.StringIO]): location of the tracking data .txt
        file channels (dict): dictionary with for all timestamps the order of which
        players are referred to in the raw tracking data
        pitch_dimensions (list): x and y dimensions of the pitch in meters
        verbose (bool, optional): whether to print information about the progress in the
        terminal. Defaults to True.

    Returns:
        pd.DataFrame: tracking data of the match in a pd dataframe
    """

    if isinstance(tracking_data_loc, str):
        file = open(tracking_data_loc)
        lines = file.readlines()
        file.close()
    else:
        lines = tracking_data_loc.readlines()

    size_lines = len(lines)
    data = {
        "timestamp": [np.nan] * size_lines,
        "ball_x": [np.nan] * size_lines,
        "ball_y": [np.nan] * size_lines,
        "ball_z": [np.nan] * size_lines,
        "ball_status": [None] * size_lines,
        "ball_posession": [None] * size_lines,
    }

    if verbose:
        print("Writing lines to dataframe:")
        lines = tqdm(lines)

    for idx, line in enumerate(lines):

        timestamp, players_info, ball_info = line.split(":")
        timestamp = _to_int(timestamp)
        data["timestamp"][idx] = timestamp

        channel = channels.loc[
            (channels["start"] <= timestamp) & (channels["end"] >= timestamp),
            "ids",
        ].iloc[0]

        players = players_info.split(";")
        for i, player in enumerate(players):
            x, y = player.split(",")
            team = channel[i].split("_")[0]
            shirt_num = channel[i].split("_")[1]
            data = _add_player_tracking_data_to_dict(team, shirt_num, x, y, data, idx)

        x, y = ball_info.split(",")
        data = _add_ball_data_to_dict(x, y, np.nan, None, None, data, idx)

    df = pd.DataFrame(data)
    df["ball_status"] = np.where(pd.isnull(df["ball_x"]), "dead", "alive")

    for col in [x for x in df.columns if "_x" in x]:
        df[col] = df[col] * pitch_dimensions[0] - (pitch_dimensions[0] / 2)

    for col in [x for x in df.columns if "_y" in x]:
        df[col] = df[col] * pitch_dimensions[1] - (pitch_dimensions[1] / 2)

    df = _insert_missing_rows(df, "timestamp")

    return df



import pandas as pd


def _normalize_playing_direction_tracking(
    td: pd.DataFrame, periods: pd.DataFrame
) -> pd.DataFrame:
    """Function to represent the home team as playing from left to right for the
    full match, and the away team from right to left.

    Args:
        td (pd.DataFrame): tracking data of the match
        periods (pd.DataFrame): description of the start and end frames of the periods

    Returns:
        pd.DataFrame: tracking data of the match, normalized in x and y direction in
        such a way that the home team is represented of playing from left to right
        for the full match.
    """

    home_x = [x for x in td.columns if "_x" in x and "home" in x]
    all_x_y = [x for x in td.columns if "_x" in x or "_y" in x]
    for _, period_row in periods.iterrows():
        if len(td[td["timestamp"] == period_row["start_frame"]].index) > 0:

            idx_start = td[td["timestamp"] == period_row["start_frame"]].index[0]
            idx_end = td[td["timestamp"] == period_row["end_frame"]].index[0]

            if td.loc[idx_start, home_x].mean() > 0:
                td.loc[idx_start:idx_end, all_x_y] *= -1

    return td



from typing import Union

import numpy as np


def _to_int(value) -> int:
    """Function to make a integer of the value if possible, else -999

    Args:
        value (): a variable value

    Returns:
       int: integer if value can be changed to integer, else -999
    """
    try:
        value = _to_float(value)
        return int(value)
    except (TypeError, ValueError):
        return -999


def _to_float(value) -> Union[float, int]:
    """Function to make a float of the value if possible, else np.nan

    Args:
        value (): a variable value

    Returns:
        Union[float, int]: integer if value can be changed to integer, else np.nan
    """
    try:
        return float(value)
    except (TypeError, ValueError):
        return np.nan



# read version from installed package
from importlib.metadata import version

__version__ = version("databallpy")


class DataBallPyError(Exception):
    "Error class specific for databallpy"

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)



import subprocess
from functools import wraps
from typing import Tuple

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm import tqdm

from databallpy.match import Match


def requires_ffmpeg(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            subprocess.run(
                ["ffmpeg", "-version"],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except FileNotFoundError:
            raise FileNotFoundError(
                "It seems like ffmpeg is not added to your python path, please install\
                     add ffmpeg to you python path to use this code."
            )

        return func(*args, **kwargs)

    return wrapper


def plot_soccer_pitch(
    field_dimen: tuple = (106.0, 68.0),
    pitch_color: str = "mediumseagreen",
    linewidth: int = 2,
    markersize: int = 20,
) -> Tuple[plt.figure, plt.axes]:
    """A function to plot a soccer pitch
    Note: relies heavily on https://github.com/Friends-of-Tracking-Data-FoTD/
    LaurieOnTracking/blob/master/Metrica_Viz.py

    Args:
        field_dimen (tuple, optional): x and y length of pitch in meters. Defaults to
        (106.0, 68.0).
        pitch_color (str, optional): Color of the pitch. Defaults to "mediumseagreen".
        linewidth (int, optional): Width of the lines on the pitch. Defaults to 2.
        markersize (int, optional): Size of the dots on the pitch. Defaults to 20.

    Returns:
        Tuple[plt.figure, plt.axes]: figure and axes with the pitch depicted on it
    """

    fig, ax = plt.subplots(figsize=(12, 8))

    # Set pitch and line colors
    ax.set_facecolor(pitch_color)
    if pitch_color not in ["white", "w"]:
        lc = "whitesmoke"  # line color
        pc = "w"  # 'spot' colors
    else:
        lc = "k"
        pc = "k"

    # All dimensions in meters
    border_dimen = (3, 3)  # include a border arround of the field of width 3m
    half_pitch_length = field_dimen[0] / 2.0  # length of half pitch
    half_pitch_width = field_dimen[1] / 2.0  # width of half pitch

    # Soccer field dimensions are in yards, so we need to convert them to meters
    meters_per_yard = 0.9144  # unit conversion from yards to meters
    goal_line_width = 8 * meters_per_yard
    box_width = 20 * meters_per_yard
    box_length = 6 * meters_per_yard
    area_width = 44 * meters_per_yard
    area_length = 18 * meters_per_yard
    penalty_spot = 12 * meters_per_yard
    corner_radius = 1 * meters_per_yard
    D_length = 8 * meters_per_yard
    D_radius = 10 * meters_per_yard
    D_pos = 12 * meters_per_yard
    centre_circle_radius = 10 * meters_per_yard

    # Plot half way line
    ax.plot([0, 0], [-half_pitch_width, half_pitch_width], lc, linewidth=linewidth)
    ax.scatter(0.0, 0.0, marker="o", facecolor=lc, linewidth=0, s=markersize)
    # Plot center circle
    y = np.linspace(-1, 1, 150) * centre_circle_radius
    x = np.sqrt(centre_circle_radius**2 - y**2)
    ax.plot(x, y, lc, linewidth=linewidth)
    ax.plot(-x, y, lc, linewidth=linewidth)

    signs = [-1, 1]
    for s in signs:  # plots each line seperately

        # Plot pitch boundary
        ax.plot(
            [-half_pitch_length, half_pitch_length],
            [s * half_pitch_width, s * half_pitch_width],
            lc,
            linewidth=linewidth,
        )
        ax.plot(
            [s * half_pitch_length, s * half_pitch_length],
            [-half_pitch_width, half_pitch_width],
            lc,
            linewidth=linewidth,
        )

        # Goal posts & line
        ax.plot(
            [s * half_pitch_length, s * half_pitch_length],
            [-goal_line_width / 2.0, goal_line_width / 2.0],
            pc + "s",
            markersize=6 * markersize / 20.0,
            linewidth=linewidth,
        )

        # 6 yard box
        ax.plot(
            [s * half_pitch_length, s * half_pitch_length - s * box_length],
            [box_width / 2.0, box_width / 2.0],
            lc,
            linewidth=linewidth,
        )
        ax.plot(
            [s * half_pitch_length, s * half_pitch_length - s * box_length],
            [-box_width / 2.0, -box_width / 2.0],
            lc,
            linewidth=linewidth,
        )
        ax.plot(
            [
                s * half_pitch_length - s * box_length,
                s * half_pitch_length - s * box_length,
            ],
            [-box_width / 2.0, box_width / 2.0],
            lc,
            linewidth=linewidth,
        )

        # Penalty area
        ax.plot(
            [s * half_pitch_length, s * half_pitch_length - s * area_length],
            [area_width / 2.0, area_width / 2.0],
            lc,
            linewidth=linewidth,
        )
        ax.plot(
            [s * half_pitch_length, s * half_pitch_length - s * area_length],
            [-area_width / 2.0, -area_width / 2.0],
            lc,
            linewidth=linewidth,
        )
        ax.plot(
            [
                s * half_pitch_length - s * area_length,
                s * half_pitch_length - s * area_length,
            ],
            [-area_width / 2.0, area_width / 2.0],
            lc,
            linewidth=linewidth,
        )

        # Penalty spot
        ax.scatter(
            s * half_pitch_length - s * penalty_spot,
            0.0,
            marker="o",
            facecolor=lc,
            linewidth=0,
            s=markersize,
        )

        # Corner flags
        y = np.linspace(0, 1, 50) * corner_radius
        x = np.sqrt(corner_radius**2 - y**2)
        ax.plot(
            s * half_pitch_length - s * x,
            -half_pitch_width + y,
            lc,
            linewidth=linewidth,
        )
        ax.plot(
            s * half_pitch_length - s * x, half_pitch_width - y, lc, linewidth=linewidth
        )

        # Draw the half circles by the box: the D
        y = (
            np.linspace(-1, 1, 50) * D_length
        )  # D_length is the chord of the circle that defines the D
        x = np.sqrt(D_radius**2 - y**2) + D_pos
        ax.plot(s * half_pitch_length - s * x, y, lc, linewidth=linewidth)

    # remove axis labels and ticks
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_xticks([])
    ax.set_yticks([])
    # set axis limits
    xmax = field_dimen[0] / 2.0 + border_dimen[0]
    ymax = field_dimen[1] / 2.0 + border_dimen[1]
    ax.set_xlim([-xmax, xmax])
    ax.set_ylim([-ymax, ymax])
    ax.set_axisbelow(True)

    return fig, ax


def plot_events(
    match: Match,
    events: list = [],
    outcome: int = None,
    player_ids: list = [],
    team_id: int = None,
    pitch_color: str = "mediumseagreen",
    color_by_col: str = None,
    team_colors: list = ["orange", "red"],
    title: str = None,
) -> Tuple[plt.figure, plt.axes]:
    """Function to plot the locations of specific events

    Args:
        match (Match): All information about a match
        events (list, optional): Filter of events you want to plot, if empty,
        all events are plotted. Defaults to [].
        outcome (int, optional): Filter if the event should have a succesfull
        outcome (1) or not (0), if None, all outcomes are included. Defaults to None.
        player_ids (list, optional): Filter for what players to include, if empty, all
        players are included. Defaults to [].
        team_id (int, optional): Filter for what team to include, if None, both teams
        are included. Defaults to None.
        pitch_color (str, optional): Prefered color of the pitch. Defaults to
        "mediumseagreen".
        color_by_col (str, optional): If specified, colors of scatter is specified by
        this colom in match.event_data. Defaults to None.
        team_colors (list, optional): Colors by which the teams should be represented.
        Defaults to ["orange", "red"].
        title (str, optional): Title of the plot. Defaults to None.

    Returns:
        Tuple[plt.figure, plt.axes]: figure and axes with the pitch and events depicted
        on it.
    """
    event_data = match.event_data

    if len(events) > 0:
        assert all([x in match.event_data["event"].unique() for x in events])
        event_data = event_data.loc[event_data["event"].isin(events)]
    if outcome is not None:
        assert outcome in [0, 1]
        event_data = event_data.loc[event_data["outcome"] == outcome]
    if len(player_ids) > 0:
        assert all(x in match.event_data["player_id"].unique() for x in player_ids)
        event_data = event_data.loc[event_data["player_id"].isin(player_ids)]
    if team_id:
        assert team_id in [match.home_team_id, match.away_team_id]
        event_data = event_data.loc[event_data["team_id"] == team_id]

    if len(event_data) == 0:
        print(
            "No events could be found that match your\
            requirements, please try again."
        )
        return None, None
    else:
        print(f"Found {len(event_data)} matching events")

    fig, ax = plot_soccer_pitch(
        field_dimen=match.pitch_dimensions, pitch_color=pitch_color
    )
    if title:
        ax.set_title(title)

    # Set match name
    ax.text(
        match.pitch_dimensions[0] / -2.0 + 2,
        match.pitch_dimensions[1] / 2.0 + 1.0,
        match.home_team_name,
        fontsize=14,
        c=team_colors[0],
        zorder=2.5,
    )
    ax.text(
        match.pitch_dimensions[0] / 2.0 - 15,
        match.pitch_dimensions[1] / 2.0 + 1.0,
        match.away_team_name,
        fontsize=14,
        c=team_colors[1],
        zorder=2.5,
    )

    # Check if color_by_col is specified and is a valid column name
    if color_by_col:
        assert color_by_col in match.event_data.columns

        # Color events by team if the specified column is "team_id"
        if color_by_col == "team_id":
            for id, c, team_name in zip(
                [match.home_team_id, match.away_team_id],
                team_colors,
                [match.home_team_name, match.away_team_name],
            ):
                temp_events = event_data[event_data[color_by_col] == id]
                ax.scatter(
                    temp_events["start_x"],
                    temp_events["start_y"],
                    marker="x",
                    label=team_name,
                    c=c,
                    zorder=2.5,
                )

        # Color events by unique values in the specified column
        else:
            for value in event_data[color_by_col].unique():
                temp_events = event_data[event_data[color_by_col] == value]
                ax.scatter(
                    temp_events["start_x"],
                    temp_events["start_y"],
                    marker="x",
                    label=value,
                    zorder=2.5,
                )

        # Add legend to the plot
        plt.legend(loc="upper center")

    # If color_by_col is not specified, color events using default settings
    else:
        ax.scatter(event_data["start_x"], event_data["start_y"], marker="x", zorder=2.5)

    return fig, ax


@requires_ffmpeg
def save_match_clip(
    match: Match,
    start_idx: int,
    end_idx: int,
    save_folder: str,
    *,
    title: str = "test_clip",
    team_colors: list = ["green", "red"],
    pitch_color: str = "mediumseagreen",
    events: list = [],
    variable_of_interest: pd.Series = None,
):
    """Function to save a subset of a match clip of the tracking data.

    Note that making animation is build with FFMPEG. You need to have FFMPEG installed
    on you device before being able to use this function.

    Args:
        match (Match): Match with tracking data and ohter info of the match.
        start_idx (int): Start index of what to save of the match.tracking_data df.
        end_idx (int): End index of what to save of the match.tracking_data df.
        save_folder (str): Location where to save the clip.
        title (str, optional): Title of the clip. Defaults to "test_clip".
        team_colors (list, optional): Colors of the home and away team. Defaults to
        ["green", "red"].
        pitch_color (str, optional): Color of the pitch. Defaults to "mediumseagreen".
        events (list, optional): What events should be plotted as well. Defaults to [].
        variable_of_interest (pd.Series, optional): Variable you want to have plotted
        in the clip, this is a pd.Series that should have the same index
        (start_idx:end_idx) as the tracking data that will be plotted. Defaults to None.
    """

    td = match.tracking_data.loc[start_idx:end_idx]
    td_ht = td[[x for x in match.home_players_column_ids if "_x" in x or "_y" in x]]
    td_at = td[[x for x in match.away_players_column_ids if "_x" in x or "_y" in x]]

    if variable_of_interest is not None:
        assert (
            variable_of_interest.index == td.index
        ).all(), (
            "Index of variable of interest and of the tracking data should be alike!"
        )

    if len(events) > 0:
        assert (
            "event" in match.tracking_data.columns
        ), "No event column found in match.tracking_data.columns, did you synchronize\
            event and tracking data?"

    animation_metadata = {
        "title": title,
        "artist": "Matplotlib",
        "comment": "Made with databallpy",
    }
    writer = animation.FFMpegWriter(fps=match.frame_rate, metadata=animation_metadata)
    video_loc = f"{save_folder}/{title}.mp4"

    fig, ax = plot_soccer_pitch(
        field_dimen=match.pitch_dimensions, pitch_color=pitch_color
    )

    # Set match name, non variable over time
    ax.text(
        match.pitch_dimensions[0] / -2.0 + 2,
        match.pitch_dimensions[1] / 2.0 + 1.0,
        match.home_team_name,
        fontsize=14,
        color=team_colors[0],
    )
    ax.text(
        match.pitch_dimensions[0] / 2.0 - 15,
        match.pitch_dimensions[1] / 2.0 + 1.0,
        match.away_team_name,
        fontsize=14,
        color=team_colors[1],
    )

    # Generate movie with variable info
    with writer.saving(fig, video_loc, 100):
        print("Making match clip...")
        for _, idx in enumerate(tqdm(td.index)):

            variable_fig_objs = []

            # Scatter plot the teams
            for td_team, c in zip([td_ht.loc[idx], td_at.loc[idx]], team_colors):
                x_cols = [x for x in td_team.index if x[-2:] == "_x"]
                y_cols = [y for y in td_team.index if y[-2:] == "_y"]
                fig_obj = ax.scatter(
                    td_team[x_cols], td_team[y_cols], c=c, alpha=0.7, s=90, zorder=2.5
                )
                variable_fig_objs.append(fig_obj)

                # Add shirt number to every dot
                for x, y in zip(x_cols, y_cols):
                    if pd.isnull(td_team[x]):
                        # Player not on the pitch currently
                        continue

                    # Slightly different place needed if the number has a len of 2
                    correction = 0.5 if len(x.split("_")[1]) == 1 else 0.8
                    fig_obj = ax.text(
                        td_team[x] - correction,
                        td_team[y] - 0.5,
                        x.split("_")[1],  # the shirt number
                        fontsize=9,
                        c="white",
                        zorder=3.0,
                    )
                    variable_fig_objs.append(fig_obj)

            # Plot the ball
            fig_obj = ax.scatter(
                td.loc[idx, "ball_x"], td.loc[idx, "ball_y"], c="black"
            )
            variable_fig_objs.append(fig_obj)

            # Add time info
            fig_obj = ax.text(
                -20.5,
                match.pitch_dimensions[1] / 2.0 + 1.0,
                td.loc[idx, "matchtime_td"],
                c="k",
                fontsize=14,
            )
            variable_fig_objs.append(fig_obj)

            # Add variable of interest
            if variable_of_interest is not None:
                fig_obj = ax.text(
                    -7,
                    match.pitch_dimensions[1] / 2.0 + 1.0,
                    str(variable_of_interest[idx]),
                    fontsize=14,
                )
                variable_fig_objs.append(fig_obj)

            # Add events
            # Note: this should be last in this function since it assumes that all
            # other info is already plotted in the axes
            if len(events) > 0:
                if td.loc[idx, "event"] in events:
                    event = (
                        match.event_data[
                            match.event_data["event_id"] == td.loc[idx, "event_id"]
                        ]
                        .iloc[0]
                        .T
                    )

                    player_name = event["player_name"]
                    event_name = event["event"]

                    # Add event text
                    fig_obj = ax.text(
                        5,
                        match.pitch_dimensions[1] / 2.0 + 1,
                        f"{player_name}: {event_name}",
                        fontsize=14,
                    )
                    variable_fig_objs.append(fig_obj)

                    # Highligh location on the pitch of the event
                    fig_obj = ax.scatter(
                        event["start_x"],
                        event["start_y"],
                        color="red",
                        marker="x",
                        s=50,
                    )
                    variable_fig_objs.append(fig_obj)

                    # Grap frame match.frame_rate times to 'pause' the video at
                    # this moment
                    for _ in range(match.frame_rate):
                        writer.grab_frame()

            # Save current frame
            writer.grab_frame()

            # Delete all variable axis objects
            for fig_obj in variable_fig_objs:
                fig_obj.remove()

    # Close figure
    plt.clf()
    plt.close(fig)



import numpy as np
import pandas as pd


def _add_periods_to_tracking_data(
    timestamps: pd.Series, periods_frames: pd.DataFrame
) -> pd.Series:
    """Function to add periods

    Args:
        tracking_data (pd.DataFrame): _description_
        periods_frames (pd.DataFrame): _description_

    Returns:
        pd.DataFrame: _description_
    """
    period_conditions = [
        (timestamps <= periods_frames.loc[0, "end_frame"]),
        (timestamps > periods_frames.loc[0, "end_frame"])
        & (timestamps < periods_frames.loc[1, "start_frame"]),
        (timestamps >= periods_frames.loc[1, "start_frame"])
        & (timestamps <= periods_frames.loc[1, "end_frame"]),
        (timestamps > periods_frames.loc[1, "end_frame"])
        & (timestamps < periods_frames.loc[2, "start_frame"]),
        (timestamps >= periods_frames.loc[2, "start_frame"])
        & (timestamps < periods_frames.loc[2, "end_frame"]),
        (timestamps > periods_frames.loc[2, "end_frame"])
        & (timestamps < periods_frames.loc[3, "start_frame"]),
        (timestamps >= periods_frames.loc[3, "start_frame"])
        & (timestamps < periods_frames.loc[3, "end_frame"]),
        (timestamps > periods_frames.loc[3, "end_frame"])
        & (timestamps < periods_frames.loc[4, "start_frame"]),
        (timestamps > periods_frames.loc[4, "start_frame"]),
    ]
    period_values = [
        1,
        -999,
        2,
        -999,
        3,
        -999,
        4,
        -999,
        5,
    ]

    return np.select(period_conditions, period_values).astype("int64")



def _add_ball_data_to_dict(
    ball_x: str,
    ball_y: str,
    ball_z: str,
    posession: str,
    status: str,
    data: dict,
    idx: str,
) -> dict:
    """Function that adds the data of the ball to the data dict for one frame

    Args:
        ball_info (str): containing data from the ball
        data (dict): data dictionary to write results to
        idx (int): indicates position in data dictionary

    Returns:
        dict: contains all tracking data
    """

    data["ball_x"][idx] = float(ball_x)
    data["ball_y"][idx] = float(ball_y)
    data["ball_z"][idx] = float(ball_z)
    data["ball_posession"][idx] = posession
    data["ball_status"][idx] = status

    return data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['databallpy',
 'databallpy.load_data',
 'databallpy.load_data.event_data',
 'databallpy.load_data.tracking_data']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.11.1,<5.0.0',
 'lxml>=4.9.2,<5.0.0',
 'matplotlib>=3.6.3,<4.0.0',
 'numpy>=1.24.1,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'requests>=2.28.2,<3.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'databallpy',
    'version': '0.2.1',
    'description': 'A package for loading, preprocessing, vizualising and synchronizing soccere event aand tracking data.',
    'long_description': '# databallpy\n\nA package for loading, preprocessing, vizualising and synchronizing soccer event- and tracking data.\n\nThis package is developed to create a standardized way to analyse soccer matches using both event- and tracking data. Other packages, like [kloppy](https://github.com/PySport/kloppy) and [floodlight](https://github.com/floodlight-sports/floodlight), already standardize the import of data sources. The current package goes a step further in combining different data streams from the same match. In this case, the `Match` object combines information from the event and tracking data.\n\n### Changelog for version 0.2.0\n\n- Added parser for Metrica, including an open dataset\n- Added functionality to synchronize tracking and event data\n- Added functionality to plot events\n- Fixed bug, now both tracking and event data are normalized in direction\n- Fixed unexpected behaviour, all date related objects are now datetime objects\n\n### Planned changes\n\n- Adding different filters to filter the tracking data\n- Make the `Match` object more accesable if you don\'t have tracking or event data\n- Adding features to quantify pressure, ball possession, etc. (if you have any ideas/requests, please open an issue!)\n- Adding expected passing and goals models\n\n## Installation\n\n```bash\n$ pip install databallpy\n```\n\n## Usage\n\nThe package is centered around the `Match` object. A `Match` has tracking data, event data metadata about the match.\nFor a more elaborate example, see the [example file](https://databallpy.readthedocs.io/en/latest/example.html)\n\n```console\n$ from databallpy.match import get_match, get_open_match\n$\n$ match = get_match(\n$   tracking_data_loc="../data/tracking_data.dat",\n$   tracking_metadata_loc="../data/tracking_metadata.xml",\n$   tracking_data_provider="tracab"\n$   event_data_loc="../data/event_data_f24.xml",\n$   event_metadata_loc="../data/event_metadata_f7.xml",\n$   event_data_provider="opta",\n$ )\n$\n$ # or to load an open metrica dataset of tracking and event data\n$ match = get_open_match()\n$\n$ match.home_team_name # the team name of the home playing team\n$ match.away_players # pandas dataframe with the names, ids, shirt numbers and positions of the away team\n$ match.tracking_data # pandas dataframe with tracking data of the match\n$ match.event_data # pandas dataframe with event data of the match\n```\n\nSee [the documentation](https://databallpy.readthedocs.io/en/latest/autoapi/databallpy/match/index.html) of the `Match` object for more options. Note that this package is developed to combine event and tracking data, therefore both datastreams are necessary to create a `Match` object.\n\n## Synchronization of tracking and event data\n\nTracking and event data is often poorly synchronized. For instance, when taking the event data of Opta and tracking data of Tracab, and take the timing of the events to synchronize tracking and event data we get something like this:\n\n\nAs you can see, the timing (and placing) of the events do not correspond good with the tracking data locations. Using the methodology of [this](https://kwiatkowski.io/sync.soccer) article, this package is able to synchronize tracking and event data using the Needleman-Wunsch algorithm. \n\nAfter running the following command, the events are better synchronized to the tracking data:\n\n```batch\n$ match.synchronise_tracking_and_event_data()\n```\n\n\n\n\n## Documentation\n\nThe official documentation can be found [here](https://databallpy.readthedocs.io/en/latest/autoapi/databallpy/index.html).\n\n## Providers\n\nFor now we limited providers. We are planning on adding more providers later on.\n\nEvent data providers:\n- Opta\n- Metrica\n\nTracking data providers:\n- Tracab\n- Metrica\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`databallpy` was created by Alexander Oonk & Daan Grob. It is licensed under the terms of the MIT license.\n\n## Similar projects\n\nAlthough we think this package helps when starting to analyse soccer data, other packages may be better suited for your specific needs. Make sure to check out the following packages as well:\n- [kloppy](https://github.com/PySport/kloppy)\n- [floodlight](https://github.com/floodlight-sports/floodlight)\n- [codeball](https://github.com/metrica-sports/codeball)\n\nAnd for a more specific toturials on how to get started with soccer data"\n- [Friends of Tracking](https://github.com/Friends-of-Tracking-Data-FoTD)\n\n\n\n',
    'author': 'Alexander Oonk',
    'author_email': 'alexanderoonk26@gmail.com',
    'maintainer': 'Alexander Oonk',
    'maintainer_email': 'alexanderoonk26@gmail.com',
    'url': 'https://pypi.org/project/databallpy/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd


def _insert_missing_rows(df: pd.DataFrame, col: str) -> pd.DataFrame:
    """Functions that inserts missing rows based on gaps in timestamp

    Args:
        df (pd.DataFrame): containing tracking data
        col (str): name of column containing timestamps

    Returns:
        pd.DataFrame: contains tracking data with inserted missing rows
    """
    assert (
        col in df.columns
    ), f"Calculations are based on {col} column, which is not in the df"

    missing = np.where(df[col].diff() > 1)[0]
    for start_missing in missing:
        n_missing = int(df[col].diff()[start_missing] - 1)
        start_timestamp = df.loc[start_missing, col] - n_missing
        to_add_data = {
            col: list(
                np.arange(start_timestamp, start_timestamp + n_missing, dtype=int)
            )
        }
        to_add_df = pd.DataFrame(to_add_data)
        df = pd.concat((df, to_add_df)).sort_values(by=col)

    df.reset_index(drop=True, inplace=True)

    return df



import numpy as np
import pandas as pd

from databallpy.load_data.metadata import Metadata


def _to_matchtime(secs: int, max_m: int, start_m: int) -> str:
    """Transforms the number of seconds into matchtime format

    Args:
        s (int): number of seconds since period started
        max_m (int): max number of minutes the period can last
        start_m (int): start of the period in minutes

    Returns:
        str: the time in matchtime format
    """
    seconds = str(secs % 60)
    if len(seconds) == 1:
        seconds = "0" + str(seconds)

    minutes = str(secs // 60 + start_m)
    if len(minutes) == 1:
        minutes = "0" + str(minutes)

    if int(minutes) < max_m:
        time_string = minutes + ":" + seconds
    else:
        max_time = str(max_m) + ":00"
        minutes_extra = str(int(minutes) - max_m)
        time_string = max_time + "+" + minutes_extra + ":" + seconds

    return time_string


def _get_matchtime(
    timestamp_column: pd.Series, period_column: pd.Series, metadata: Metadata
) -> list:
    """Gives a list with time in the matchtime format based
    on the original timestamps and framerate

    Args:
        timestamp_column (pd.Series): containing the timestamps from tracking data
        dataframe
        period_column (pd.Series): containing the period for every frame
        metadata (Metadata): metadata including framerate and
        information on start and end of periods

    Returns:
        list: for every frame the match time.
    """
    frame_rate = metadata.frame_rate
    periods_frames = metadata.periods_frames

    period_start_dict = dict(
        zip(periods_frames["period"], periods_frames["start_frame"])
    )

    n_frames_period = dict(
        zip(
            periods_frames["period"],
            periods_frames["end_frame"] - periods_frames["start_frame"],
        )
    )

    rel_timestamp = np.array(
        [
            x - period_start_dict[p] if p > 0 else -999 * frame_rate
            for x, p in zip(timestamp_column.values, period_column.values)
        ]
    )
    seconds = rel_timestamp // frame_rate
    df = pd.DataFrame(
        {
            "seconds": seconds,
            "period": period_column.values,
        }
    )
    start_m_dict = {1: 0, 2: 45, 3: 90, 4: 105}
    max_m_dict = {1: 45, 2: 90, 3: 105, 4: 120}

    matchtime_list = []
    for p in [1, 2, 3, 4]:
        frame_end_current_p = periods_frames.loc[
            periods_frames["period"] == p, "end_frame"
        ].iloc[0]
        frame_start_next_p = periods_frames.loc[
            periods_frames["period"] == p + 1, "start_frame"
        ].iloc[0]
        if frame_start_next_p > 0 and frame_end_current_p > 0:
            n_frames_break = frame_start_next_p - frame_end_current_p - 1
        else:
            n_frames_break = 0
        matchtime_list_period = []
        for seconds in df[df["period"] == p]["seconds"].unique():
            matchtime_list_period.extend(
                [_to_matchtime(int(seconds), max_m_dict[p], start_m_dict[p])]
                * frame_rate
            )
        matchtime_list_period = matchtime_list_period[: n_frames_period[p] + 1]
        matchtime_list_period.extend(["Break"] * n_frames_break)
        matchtime_list.extend(matchtime_list_period)

    for _ in df[df["period"] == 5]["seconds"].unique():
        matchtime_list.extend(["Penalty Shootout"] * frame_rate)

    matchtime_list = matchtime_list[: len(df)]

    return matchtime_list
